# This file is part of the pyMOR project (https://www.pymor.org).
# Copyright 2013-2021 pyMOR developers and contributors. All rights reserved.
# License: BSD 2-Clause License (https://opensource.org/licenses/BSD-2-Clause)

import numpy as np
from numpy.random import uniform

from pymor.algorithms.randrangefinder import rrf, adaptive_rrf
from pymor.operators.numpy import NumpyMatrixOperator
from pymor.operators.constructions import VectorArrayOperator


np.random.seed(0)
A = uniform(low=-1.0, high=1.0, size=(100, 100))
A = A.dot(A.T)
range_product = NumpyMatrixOperator(A)

np.random.seed(1)
A = uniform(low=-1.0, high=1.0, size=(10, 10))
A = A.dot(A.T)
source_product = NumpyMatrixOperator(A)

B = range_product.range.random(10, seed=10)
op = VectorArrayOperator(B)

C = range_product.range.random(10, seed=11)+1j*range_product.range.random(10, seed=12)
op_complex = VectorArrayOperator(C)


def test_rrf():
    Q = rrf(op, source_product, range_product)
    assert Q in op.range
    assert len(Q) == 8

    Q = rrf(op_complex, iscomplex=True)
    assert np.iscomplexobj(Q.to_numpy())
    assert Q in op.range
    assert len(Q) == 8


def test_adaptive_rrf():
    B = adaptive_rrf(op, source_product, range_product)
    assert B in op.range

    B = adaptive_rrf(op_complex, iscomplex=True)
    assert np.iscomplexobj(B.to_numpy())
    assert B in op.range



# This file is part of the pyMOR project (https://www.pymor.org).
# Copyright 2013-2021 pyMOR developers and contributors. All rights reserved.
# License: BSD 2-Clause License (https://opensource.org/licenses/BSD-2-Clause)

import numpy as np
import scipy.linalg as spla
import scipy.sparse as sps

from pymor.algorithms.to_matrix import to_matrix
from pymor.operators.block import BlockOperator, BlockDiagonalOperator
from pymor.operators.constructions import (AdjointOperator, ComponentProjectionOperator, IdentityOperator,
                                           LowRankOperator, LowRankUpdatedOperator, VectorArrayOperator, ZeroOperator)
from pymor.operators.numpy import NumpyMatrixOperator
from pymor.vectorarrays.numpy import NumpyVectorSpace


def assert_type_and_allclose(A, Aop, default_format):
    if default_format == 'dense':
        assert isinstance(to_matrix(Aop), np.ndarray)
        assert np.allclose(A, to_matrix(Aop))
    elif default_format == 'sparse':
        assert sps.issparse(to_matrix(Aop))
        assert np.allclose(A, to_matrix(Aop).toarray())
    else:
        assert getattr(sps, 'isspmatrix_' + default_format)(to_matrix(Aop))
        assert np.allclose(A, to_matrix(Aop).toarray())

    assert isinstance(to_matrix(Aop, format='dense'), np.ndarray)
    assert np.allclose(A, to_matrix(Aop, format='dense'))

    assert sps.isspmatrix_csr(to_matrix(Aop, format='csr'))
    assert np.allclose(A, to_matrix(Aop, format='csr').toarray())


def test_to_matrix_NumpyMatrixOperator():
    np.random.seed(0)
    A = np.random.randn(2, 2)

    Aop = NumpyMatrixOperator(A)
    assert_type_and_allclose(A, Aop, 'dense')

    Aop = NumpyMatrixOperator(sps.csc_matrix(A))
    assert_type_and_allclose(A, Aop, 'csc')


def test_to_matrix_BlockOperator():
    np.random.seed(0)
    A11 = np.random.randn(2, 2)
    A12 = np.random.randn(2, 3)
    A21 = np.random.randn(3, 2)
    A22 = np.random.randn(3, 3)
    B = np.asarray(np.bmat([[A11, A12], [A21, A22]]))

    A11op = NumpyMatrixOperator(A11)
    A12op = NumpyMatrixOperator(A12)
    A21op = NumpyMatrixOperator(A21)
    A22op = NumpyMatrixOperator(A22)
    Bop = BlockOperator([[A11op, A12op], [A21op, A22op]])
    assert_type_and_allclose(B, Bop, 'dense')

    A11op = NumpyMatrixOperator(sps.csc_matrix(A11))
    A12op = NumpyMatrixOperator(A12)
    A21op = NumpyMatrixOperator(A21)
    A22op = NumpyMatrixOperator(A22)
    Bop = BlockOperator([[A11op, A12op], [A21op, A22op]])
    assert_type_and_allclose(B, Bop, 'sparse')


def test_to_matrix_BlockDiagonalOperator():
    np.random.seed(0)
    A1 = np.random.randn(2, 2)
    A2 = np.random.randn(3, 3)
    B = np.asarray(np.bmat([[A1, np.zeros((2, 3))],
                            [np.zeros((3, 2)), A2]]))

    A1op = NumpyMatrixOperator(A1)
    A2op = NumpyMatrixOperator(A2)
    Bop = BlockDiagonalOperator([A1op, A2op])
    assert_type_and_allclose(B, Bop, 'sparse')

    A1op = NumpyMatrixOperator(sps.csc_matrix(A1))
    A2op = NumpyMatrixOperator(A2)
    Bop = BlockDiagonalOperator([A1op, A2op])
    assert_type_and_allclose(B, Bop, 'sparse')


def test_to_matrix_AdjointOperator():
    np.random.seed(0)
    A = np.random.randn(2, 2)
    S = np.random.randn(2, 2)
    S = S.dot(S.T)
    R = np.random.randn(2, 2)
    R = R.dot(R.T)

    Aop = NumpyMatrixOperator(A)
    Aadj = AdjointOperator(Aop)
    assert_type_and_allclose(A.T, Aadj, 'dense')

    Aop = NumpyMatrixOperator(sps.csc_matrix(A))
    Aadj = AdjointOperator(Aop)
    assert_type_and_allclose(A.T, Aadj, 'sparse')

    Aop = NumpyMatrixOperator(A)
    Sop = NumpyMatrixOperator(S)
    Rop = NumpyMatrixOperator(R)
    Aadj = AdjointOperator(Aop, source_product=Sop, range_product=Rop)
    assert_type_and_allclose(spla.solve(S, A.T.dot(R)), Aadj, 'dense')

    Aop = NumpyMatrixOperator(sps.csc_matrix(A))
    Sop = NumpyMatrixOperator(S)
    Rop = NumpyMatrixOperator(R)
    Aadj = AdjointOperator(Aop, source_product=Sop, range_product=Rop)
    assert_type_and_allclose(spla.solve(S, A.T.dot(R)), Aadj, 'dense')

    Aop = NumpyMatrixOperator(A)
    Sop = NumpyMatrixOperator(S)
    Rop = NumpyMatrixOperator(sps.csc_matrix(R))
    Aadj = AdjointOperator(Aop, source_product=Sop, range_product=Rop)
    assert_type_and_allclose(spla.solve(S, A.T.dot(R)), Aadj, 'dense')

    Aop = NumpyMatrixOperator(sps.csc_matrix(A))
    Sop = NumpyMatrixOperator(sps.csc_matrix(S))
    Rop = NumpyMatrixOperator(sps.csc_matrix(R))
    Aadj = AdjointOperator(Aop, source_product=Sop, range_product=Rop)
    assert_type_and_allclose(spla.solve(S, A.T.dot(R)), Aadj, 'sparse')


def test_to_matrix_ComponentProjectionOperator():
    dofs = np.array([0, 1, 2, 4, 8])
    n = 10
    A = np.zeros((len(dofs), n))
    A[range(len(dofs)), dofs] = 1

    source = NumpyVectorSpace(n)
    Aop = ComponentProjectionOperator(dofs, source)
    assert_type_and_allclose(A, Aop, 'sparse')


def test_to_matrix_ConcatenationOperator():
    np.random.seed(0)
    A = np.random.randn(2, 3)
    B = np.random.randn(3, 4)
    C = A.dot(B)

    Aop = NumpyMatrixOperator(A)
    Bop = NumpyMatrixOperator(B)
    Cop = Aop @ Bop
    assert_type_and_allclose(C, Cop, 'dense')

    Aop = NumpyMatrixOperator(sps.csc_matrix(A))
    Bop = NumpyMatrixOperator(B)
    Cop = Aop @ Bop
    assert_type_and_allclose(C, Cop, 'dense')

    Aop = NumpyMatrixOperator(A)
    Bop = NumpyMatrixOperator(sps.csc_matrix(B))
    Cop = Aop @ Bop
    assert_type_and_allclose(C, Cop, 'dense')

    Aop = NumpyMatrixOperator(sps.csc_matrix(A))
    Bop = NumpyMatrixOperator(sps.csc_matrix(B))
    Cop = Aop @ Bop
    assert_type_and_allclose(A, Aop, 'sparse')


def test_to_matrix_IdentityOperator():
    n = 3
    I = np.eye(n)

    Iop = IdentityOperator(NumpyVectorSpace(n))
    assert_type_and_allclose(I, Iop, 'sparse')


def test_to_matrix_LincombOperator():
    np.random.seed(0)
    A = np.random.randn(3, 3)
    B = np.random.randn(3, 2)
    a = np.random.randn()
    b = np.random.randn()
    C = a * A + b * B.dot(B.T)

    Aop = NumpyMatrixOperator(A)
    Bop = NumpyMatrixOperator(B)
    Cop = Aop * a + (Bop @ Bop.H) * b
    assert_type_and_allclose(C, Cop, 'dense')

    Aop = NumpyMatrixOperator(sps.csc_matrix(A))
    Bop = NumpyMatrixOperator(B)
    Cop = Aop * a + (Bop @ Bop.H) * b
    assert_type_and_allclose(C, Cop, 'dense')

    Aop = NumpyMatrixOperator(A)
    Bop = NumpyMatrixOperator(sps.csc_matrix(B))
    Cop = Aop * a + (Bop @ Bop.H) * b
    assert_type_and_allclose(C, Cop, 'dense')

    Aop = NumpyMatrixOperator(sps.csc_matrix(A))
    Bop = NumpyMatrixOperator(sps.csc_matrix(B))
    Cop = Aop * a + (Bop @ Bop.H) * b
    assert_type_and_allclose(C, Cop, 'sparse')


def test_to_matrix_LowRankOperator():
    np.random.seed(0)
    m = 6
    n = 5
    r = 2
    L = np.random.randn(m, r)
    Lva = NumpyVectorSpace.make_array(L.T)
    C = np.random.randn(r, r)
    R = np.random.randn(n, r)
    Rva = NumpyVectorSpace.make_array(R.T)

    LR = LowRankOperator(Lva, C, Rva)
    assert_type_and_allclose(L @ C @ R.T, LR, 'dense')

    LR = LowRankOperator(Lva, C, Rva, inverted=True)
    assert_type_and_allclose(L @ spla.solve(C, R.T), LR, 'dense')


def test_to_matrix_LowRankUpdatedOperator():
    np.random.seed(0)
    m = 6
    n = 5
    r = 2
    A = np.random.randn(m, n)
    Aop = NumpyMatrixOperator(A)
    L = np.random.randn(m, r)
    Lva = NumpyVectorSpace.make_array(L.T)
    C = np.random.randn(r, r)
    R = np.random.randn(n, r)
    Rva = NumpyVectorSpace.make_array(R.T)
    LR = LowRankOperator(Lva, C, Rva)

    op = LowRankUpdatedOperator(Aop, LR, 1, 1)
    assert_type_and_allclose(A + L @ C @ R.T, op, 'dense')


def test_to_matrix_VectorArrayOperator():
    np.random.seed(0)
    V = np.random.randn(10, 2)

    Vva = NumpyVectorSpace.make_array(V.T)
    Vop = VectorArrayOperator(Vva)
    assert_type_and_allclose(V, Vop, 'dense')

    Vop = VectorArrayOperator(Vva, adjoint=True)
    assert_type_and_allclose(V.T, Vop, 'dense')


def test_to_matrix_ZeroOperator():
    n = 3
    m = 4
    Z = np.zeros((n, m))

    Zop = ZeroOperator(NumpyVectorSpace(n), NumpyVectorSpace(m))
    assert_type_and_allclose(Z, Zop, 'sparse')



# This file is part of the pyMOR project (https://www.pymor.org).
# Copyright 2013-2021 pyMOR developers and contributors. All rights reserved.
# License: BSD 2-Clause License (https://opensource.org/licenses/BSD-2-Clause)


from pymor.core.config import config, _PACKAGES


def test_repr():
    repr(config)


def test_entries():
    for p in _PACKAGES:
        assert hasattr(config, 'HAVE_' + p)
        assert type(getattr(config, 'HAVE_' + p)) is bool
        getattr(config, p + '_VERSION')


def test_dir():
    d = dir(config)
    for p in _PACKAGES:
        assert 'HAVE_' + p in d
        assert p + '_VERSION' in d



# This file is part of the pyMOR project (https://www.pymor.org).
# Copyright 2013-2021 pyMOR developers and contributors. All rights reserved.
# License: BSD 2-Clause License (https://opensource.org/licenses/BSD-2-Clause)

from itertools import product

import numpy as np

from pymor.algorithms.to_matrix import to_matrix
from pymor.algorithms.simplify import expand
from pymor.operators.constructions import LincombOperator, ConcatenationOperator
from pymor.operators.numpy import NumpyMatrixOperator
from pymor.parameters.functionals import ProjectionParameterFunctional


def test_expand():
    ops = [NumpyMatrixOperator(np.eye(1) * i) for i in range(8)]
    pfs = [ProjectionParameterFunctional('p', 9, i) for i in range(8)]
    prods = [o * p for o, p in zip(ops, pfs)]

    op = ((prods[0] + prods[1] + prods[2]) @ (prods[3] + prods[4] + prods[5]) @
          (prods[6] + prods[7]))

    eop = expand(op)

    assert isinstance(eop, LincombOperator)
    assert len(eop.operators) == 3 * 3 * 2
    assert all(isinstance(o, ConcatenationOperator) and len(o.operators) == 3
               for o in eop.operators)
    assert ({to_matrix(o)[0, 0] for o in eop.operators}
            == {i0 * i1 * i2 for i0, i1, i2 in product([0, 1, 2], [3, 4, 5], [6, 7])})
    assert ({frozenset(p.index for p in pf.factors) for pf in eop.coefficients}
            == {frozenset([i0, i1, i2]) for i0, i1, i2 in product([0, 1, 2], [3, 4, 5], [6, 7])})



# This file is part of the pyMOR project (https://www.pymor.org).
# Copyright 2013-2021 pyMOR developers and contributors. All rights reserved.
# License: BSD 2-Clause License (https://opensource.org/licenses/BSD-2-Clause)

import numpy as np

from pymor.algorithms.pod import pod
from pymor.algorithms.timestepping import ExplicitEulerTimeStepper
from pymor.discretizers.builtin.gui.visualizers import OnedVisualizer
from pymor.models.basic import InstationaryModel
from pymor.discretizers.builtin import OnedGrid
from pymor.operators.constructions import VectorOperator, LincombOperator
from pymor.parameters.functionals import ProjectionParameterFunctional
from pymor.reductors.basic import InstationaryRBReductor

# import wrapped classes
from pymordemos.minimal_cpp_demo.wrapper import WrappedDiffusionOperator


def discretize(n, nt, blocks):
    h = 1. / blocks
    ops = [WrappedDiffusionOperator.create(n, h * i, h * (i + 1)) for i in range(blocks)]
    pfs = [ProjectionParameterFunctional('diffusion_coefficients', blocks, i) for i in range(blocks)]
    operator = LincombOperator(ops, pfs)

    initial_data = operator.source.zeros()

    # use data property of WrappedVector to setup rhs
    # note that we cannot use the data property of ListVectorArray,
    # since ListVectorArray will always return a copy
    rhs_vec = operator.range.zeros()
    rhs_data = rhs_vec._list[0].to_numpy()
    rhs_data[:] = np.ones(len(rhs_data))
    rhs_data[0] = 0
    rhs_data[len(rhs_data) - 1] = 0
    rhs = VectorOperator(rhs_vec)

    # hack together a visualizer ...
    grid = OnedGrid(domain=(0, 1), num_intervals=n)
    visualizer = OnedVisualizer(grid)

    time_stepper = ExplicitEulerTimeStepper(nt)

    fom = InstationaryModel(T=1e-0, operator=operator, rhs=rhs, initial_data=initial_data,
                            time_stepper=time_stepper, num_values=20,
                            visualizer=visualizer, name='C++-Model')
    return fom


# discretize
fom = discretize(50, 10000, 4)
parameter_space = fom.parameters.space(0.1, 1)

# generate solution snapshots
snapshots = fom.solution_space.empty()
for mu in parameter_space.sample_uniformly(2):
    snapshots.append(fom.solve(mu))

# apply POD
reduced_basis = pod(snapshots, modes=4)[0]

# reduce the model
reductor = InstationaryRBReductor(fom, reduced_basis, check_orthonormality=True)
rom = reductor.reduce()

# stochastic error estimation
mu_max = None
err_max = -1.
for mu in parameter_space.sample_randomly(10):
    U_RB = (reductor.reconstruct(rom.solve(mu)))
    U = fom.solve(mu)
    err = np.max((U_RB-U).norm())
    if err > err_max:
        err_max = err
        mu_max = mu

# visualize maximum error solution
U_RB = (reductor.reconstruct(rom.solve(mu_max)))
U = fom.solve(mu_max)

# avoid blocking in CI run
if __name__ == '__main__':
    fom.visualize((U_RB, U), title=f'mu = {mu}', legend=('reduced', 'detailed'))
    fom.visualize((U-U_RB), title=f'mu = {mu}', legend=('error'))



# This file is part of the pyMOR project (https://www.pymor.org).
# Copyright 2013-2021 pyMOR developers and contributors. All rights reserved.
# License: BSD 2-Clause License (https://opensource.org/licenses/BSD-2-Clause)

import logging

import pytest

import pymor.core as core
from pymor.core.logger import log_levels
from pymor.operators.numpy import NumpyMatrixOperator
from pymortests.base import (runmodule,)


def test_logger():
    logger = NumpyMatrixOperator._logger
    for lvl in [getattr(logging, lvl) for lvl in ['WARN', 'ERROR', 'DEBUG', 'INFO']]:
        logger.setLevel(lvl)
        assert logger.isEnabledFor(lvl)
    for verb in ['warning', 'error', 'debug', 'info']:
        getattr(logger, verb)(f'{verb} -- logger {str(logger)}')


def test_empty_log_message():
    core.logger.getLogger('test').warning('')


def test_log_levels():
    logger = NumpyMatrixOperator._logger
    before_name = 'INFO'
    logger.setLevel(before_name)
    before = logger.level
    with log_levels({logger.name: 'DEBUG'}):
        assert 'DEBUG' == logging.getLevelName(logger.level)
        assert logger.level != before
    assert logger.level == before
    assert before_name == logging.getLevelName(logger.level)


@pytest.mark.parametrize('verb', ('info', 'error', 'fatal', 'debug', 'block', 'info2', 'info3', 'warning'))
def test_once(verb, capsys):
    logger = NumpyMatrixOperator._logger
    logger.setLevel('DEBUG')
    func = getattr(logger, f'{verb}_once')
    msg = f'{verb} -- logger {str(logger)}'
    func(msg)
    # this just clears the capture buffer
    capsys.readouterr()
    func(msg)
    second = capsys.readouterr()
    # same log call must result in no output
    assert second.out == second.err == ''


if __name__ == "__main__":
    runmodule(filename=__file__)



# This file is part of the pyMOR project (https://www.pymor.org).
# Copyright 2013-2021 pyMOR developers and contributors. All rights reserved.
# License: BSD 2-Clause License (https://opensource.org/licenses/BSD-2-Clause)

from numbers import Number

import numpy as np
from scipy.sparse import issparse

from pymor.core.base import classinstancemethod
from pymor.tools.random import get_random_state
from pymor.vectorarrays.interface import VectorArray, VectorSpace, _create_random_values


class NumpyVectorArray(VectorArray):
    """|VectorArray| implementation via |NumPy arrays|.

    This is the default |VectorArray| type used by all |Operators|
    in pyMOR's discretization toolkit. Moreover, all reduced |Operators|
    are based on |NumpyVectorArray|.

    This class is just a thin wrapper around the underlying
    |NumPy array|. Thus, while operations like
    :meth:`~pymor.vectorarrays.interface.VectorArray.axpy` or
    :meth:`~pymor.vectorarrays.interface.VectorArray.inner`
    will be quite efficient, removing or appending vectors will
    be costly.

    .. warning::
        This class is not intended to be instantiated directly. Use
        the associated :class:`VectorSpace <NumpyVectorSpace>` instead.
    """

    def __init__(self, array, space):
        self._array = array
        self.space = space
        self._refcount = [1]
        self._len = len(array)

    def to_numpy(self, ensure_copy=False):
        if ensure_copy:
            return self._array[:self._len].copy()
        if self._refcount[0] > 1:
            # While changing the returned NumPy array may change the data of `self`,
            # it should not change the data of other arrays.
            self._deep_copy()
        return self._array[:self._len]

    @property
    def _data(self):
        """Return NumPy Array view on data for hacking / interactive use."""
        return self.to_numpy()

    @property
    def real(self):
        return NumpyVectorArray(self.to_numpy().real.copy(), self.space)

    @property
    def imag(self):
        return NumpyVectorArray(self.to_numpy().imag.copy(), self.space)

    def conj(self):
        if np.isrealobj(self.to_numpy()):
            return self.copy()
        return NumpyVectorArray(np.conj(self.to_numpy()), self.space)

    def __len__(self):
        return self._len

    def __getitem__(self, ind):
        if isinstance(ind, Number) and (ind >= self._len or ind < -self._len):
            raise IndexError('VectorArray index out of range')
        return NumpyVectorArrayView(self, ind)

    def __delitem__(self, ind):
        assert self.check_ind(ind)
        if self._refcount[0] > 1:
            self._deep_copy()

        if type(ind) is slice:
            ind = set(range(*ind.indices(self._len)))
        elif not hasattr(ind, '__len__'):
            ind = {ind if 0 <= ind else self._len + ind}
        else:
            l = self._len
            ind = {i if 0 <= i else l+i for i in ind}
        remaining = sorted(set(range(len(self))) - ind)
        self._array = self._array[remaining]
        self._len = len(self._array)
        if not self._array.flags['OWNDATA']:
            self._array = self._array.copy()

    def copy(self, deep=False, *, _ind=None):
        if _ind is None and not deep:
            C = NumpyVectorArray(self._array, self.space)
            C._len = self._len
            C._refcount = self._refcount
            self._refcount[0] += 1
            return C
        else:
            new_array = self._array[:self._len] if _ind is None else self._array[_ind]
            if not new_array.flags['OWNDATA']:
                new_array = new_array.copy()
            return NumpyVectorArray(new_array, self.space)

    def append(self, other, remove_from_other=False):
        assert self.dim == other.dim
        assert not remove_from_other or (other is not self and getattr(other, 'base', None) is not self)

        if self._refcount[0] > 1:
            self._deep_copy()

        other_array = other.to_numpy()
        len_other = len(other_array)
        if len_other == 0:
            return

        if len_other <= self._array.shape[0] - self._len:
            if self._array.dtype != other_array.dtype:
                self._array = self._array.astype(np.promote_types(self._array.dtype, other_array.dtype))
            self._array[self._len:self._len + len_other] = other_array
        else:
            self._array = np.append(self._array[:self._len], other_array, axis=0)
        self._len += len_other

        if remove_from_other:
            if other.is_view:
                del other.base[other.ind]
            else:
                del other[:]

    def scal(self, alpha, *, _ind=None):
        if _ind is None:
            _ind = slice(0, self._len)
        assert isinstance(alpha, Number) \
            or isinstance(alpha, np.ndarray) and alpha.shape == (self.len_ind(_ind),)

        if self._refcount[0] > 1:
            self._deep_copy()

        if type(alpha) is np.ndarray:
            alpha = alpha[:, np.newaxis]

        alpha_type = type(alpha)
        alpha_dtype = alpha.dtype if alpha_type is np.ndarray else alpha_type
        if self._array.dtype != alpha_dtype:
            self._array = self._array.astype(np.promote_types(self._array.dtype, alpha_dtype))
        self._array[_ind] *= alpha

    def axpy(self, alpha, x, *, _ind=None):
        if _ind is None:
            _ind = slice(0, self._len)
        assert self.dim == x.dim
        assert isinstance(alpha, Number) \
            or isinstance(alpha, np.ndarray) and alpha.shape == (self.len_ind(_ind),)

        if self._refcount[0] > 1:
            self._deep_copy()

        B = x.base._array[x.ind] if x.is_view else x._array[:x._len]
        assert self.len_ind(_ind) == len(B) or len(B) == 1

        alpha_type = type(alpha)
        alpha_dtype = alpha.dtype if alpha_type is np.ndarray else alpha_type
        if self._array.dtype != alpha_dtype or self._array.dtype != B.dtype:
            dtype = np.promote_types(self._array.dtype, alpha_dtype)
            dtype = np.promote_types(dtype, B.dtype)
            self._array = self._array.astype(dtype)

        if type(alpha) is np.ndarray:
            alpha = alpha[:, np.newaxis]
        self._array[_ind] += B * alpha

    def inner(self, other, product=None, *, _ind=None):
        assert product is None or _ind is None
        if product is not None:
            return product.apply2(self, other)
        if _ind is None:
            _ind = slice(0, self._len)
        assert self.dim == other.dim

        A = self._array[_ind]
        B = other.base._array[other.ind] if other.is_view else other._array[:other._len]

        # .conj() is a no-op on non-complex data types
        return A.conj().dot(B.T)

    def pairwise_inner(self, other, product=None, *, _ind=None):
        assert product is None or _ind is None
        if product is not None:
            return product.pairwise_apply2(self, other)
        if _ind is None:
            _ind = slice(0, self._len)
        assert self.dim == other.dim

        A = self._array[_ind]
        B = other.base._array[other.ind] if other.is_view else other._array[:other._len]

        assert len(A) == len(B)

        # .conj() is a no-op on non-complex data types
        return np.sum(A.conj() * B, axis=1)

    def lincomb(self, coefficients, *, _ind=None):
        if _ind is None:
            _ind = slice(0, self._len)
        assert 1 <= coefficients.ndim <= 2

        if coefficients.ndim == 1:
            coefficients = coefficients[np.newaxis, ...]

        return NumpyVectorArray(coefficients.dot(self._array[_ind]), self.space)

    def _norm(self, *, _ind=None):
        if _ind is None:
            _ind = slice(0, self._len)
        return np.linalg.norm(self._array[_ind], axis=1)

    def _norm2(self, *, _ind=None):
        if _ind is None:
            _ind = slice(0, self._len)
        A = self._array[_ind]
        return np.sum((A * A.conj()).real, axis=1)

    def sup_norm(self, *, _ind=None):
        if self.dim == 0:
            if _ind is None:
                _ind = slice(0, self._len)
            return np.zeros(self.len_ind(_ind))
        else:
            _, max_val = self.amax(_ind=_ind)
            return max_val

    def dofs(self, dof_indices, *, _ind=None):
        if _ind is None:
            _ind = slice(0, self._len)
        assert isinstance(dof_indices, list) and (len(dof_indices) == 0 or min(dof_indices) >= 0) \
            or (isinstance(dof_indices, np.ndarray) and dof_indices.ndim == 1
                and (len(dof_indices) == 0 or np.min(dof_indices) >= 0))
        # NumPy 1.9 is quite permissive when indexing arrays of size 0, so we have to add the
        # following check:
        assert self._len > 0 \
            or (isinstance(dof_indices, list)
                and (len(dof_indices) == 0 or max(dof_indices) < self.dim)) \
            or (isinstance(dof_indices, np.ndarray) and dof_indices.ndim == 1
                and (len(dof_indices) == 0 or np.max(dof_indices) < self.dim))

        return self._array[:, dof_indices][_ind, :]

    def amax(self, *, _ind=None):
        if _ind is None:
            _ind = slice(0, self._len)
        assert self.dim > 0

        if self._array.shape[1] == 0:
            l = self.len_ind(_ind)
            return np.ones(l) * -1, np.zeros(l)

        A = np.abs(self._array[_ind])
        max_ind = np.argmax(A, axis=1)
        max_val = A[np.arange(len(A)), max_ind]
        return max_ind, max_val

    def __str__(self):
        return self._array[:self._len].__str__()

    def _format_repr(self, max_width, verbosity):
        return super()._format_repr(max_width, verbosity, override={'array': str(self._array[:self._len].__str__())})

    def __del__(self):
        self._refcount[0] -= 1

    def _deep_copy(self):
        self._array = self._array.copy()  # copy the array data
        self._refcount[0] -= 1            # decrease refcount for original array
        self._refcount = [1]              # create new reference counter

    def __add__(self, other):
        if isinstance(other, Number):
            assert other == 0
            return self.copy()
        assert self.dim == other.dim
        return NumpyVectorArray(self._array[:self._len]
                                + (other.base._array[other.ind] if other.is_view else other._array[:other._len]),
                                self.space)

    def __iadd__(self, other):
        assert self.dim == other.dim
        if self._refcount[0] > 1:
            self._deep_copy()
        other_dtype = other.base._array.dtype if other.is_view else other._array.dtype
        common_dtype = np.promote_types(self._array.dtype, other_dtype)
        if self._array.dtype != common_dtype:
            self._array = self._array.astype(common_dtype)
        self._array[:self._len] += other.base._array[other.ind] if other.is_view else other._array[:other._len]
        return self

    __radd__ = __add__

    def __sub__(self, other):
        assert self.dim == other.dim
        return NumpyVectorArray(self._array[:self._len]
                                - (other.base._array[other.ind] if other.is_view else other._array[:other._len]),
                                self.space)

    def __isub__(self, other):
        assert self.dim == other.dim
        if self._refcount[0] > 1:
            self._deep_copy()
        other_dtype = other.base._array.dtype if other.is_view else other._array.dtype
        common_dtype = np.promote_types(self._array.dtype, other_dtype)
        if self._array.dtype != common_dtype:
            self._array = self._array.astype(common_dtype)
        self._array[:self._len] -= other.base._array[other.ind] if other.is_view else other._array[:other._len]
        return self

    def __mul__(self, other):
        assert isinstance(other, Number) \
            or isinstance(other, np.ndarray) and other.shape == (len(self),)
        if type(other) is np.ndarray:
            other = other[:, np.newaxis]
        return NumpyVectorArray(self._array[:self._len] * other, self.space)

    def __imul__(self, other):
        assert isinstance(other, Number) \
            or isinstance(other, np.ndarray) and other.shape == (len(self),)
        if type(other) is np.ndarray:
            other = other[:, np.newaxis]
        if self._refcount[0] > 1:
            self._deep_copy()
        other_dtype = other.dtype if isinstance(other, np.ndarray) else type(other)
        common_dtype = np.promote_types(self._array.dtype, other_dtype)
        if self._array.dtype != common_dtype:
            self._array = self._array.astype(common_dtype)
        self._array[:self._len] *= other
        return self

    def __neg__(self):
        return NumpyVectorArray(-self._array[:self._len], self.space)


class NumpyVectorSpace(VectorSpace):
    """|VectorSpace| of |NumpyVectorArrays|.

    Parameters
    ----------
    dim
        The dimension of the vectors contained in the space.
    id
        See :attr:`~pymor.vectorarrays.interface.VectorSpace.id`.
    """

    def __init__(self, dim, id=None):
        self.dim = dim
        self.id = id

    def __eq__(self, other):
        return type(other) is type(self) and self.dim == other.dim and self.id == other.id

    def __hash__(self):
        return hash(self.dim) + hash(self.id)

    def zeros(self, count=1, reserve=0):
        assert count >= 0
        assert reserve >= 0
        va = NumpyVectorArray(np.empty((0, 0)), self)
        va._array = np.zeros((max(count, reserve), self.dim))
        va._len = count
        return va

    def full(self, value, count=1, reserve=0):
        assert count >= 0
        assert reserve >= 0
        va = NumpyVectorArray(np.empty((0, 0)), self)
        va._array = np.full((max(count, reserve), self.dim), value)
        va._len = count
        return va

    def random(self, count=1, distribution='uniform', random_state=None, seed=None, reserve=0, **kwargs):
        assert count >= 0
        assert reserve >= 0
        assert random_state is None or seed is None
        random_state = get_random_state(random_state, seed)
        va = self.zeros(count, reserve)
        va._array[:count] = _create_random_values((count, self.dim), distribution, random_state, **kwargs)
        return va

    @classinstancemethod
    def make_array(cls, obj, id=None):
        return cls._array_factory(obj, id=id)

    @make_array.instancemethod
    def make_array(self, obj):
        """:noindex:"""
        return self._array_factory(obj, space=self)

    @classinstancemethod
    def from_numpy(cls, data, id=None, ensure_copy=False):
        return cls._array_factory(data.copy() if ensure_copy else data, id=id)

    @from_numpy.instancemethod
    def from_numpy(self, data, ensure_copy=False):
        """:noindex:"""
        return self._array_factory(data.copy() if ensure_copy else data, space=self)

    @classinstancemethod
    def from_file(cls, path, key=None, single_vector=False, transpose=False, id=None):
        assert not (single_vector and transpose)
        from pymor.tools.io import load_matrix
        array = load_matrix(path, key=key)
        assert isinstance(array, np.ndarray)
        assert array.ndim <= 2
        if array.ndim == 1:
            array = array.reshape((1, -1))
        if single_vector:
            assert array.shape[0] == 1 or array.shape[1] == 1
            array = array.reshape((1, -1))
        if transpose:
            array = array.T
        return cls.make_array(array, id=id)

    @from_file.instancemethod
    def from_file(self, path, key=None, single_vector=False, transpose=False):
        """:noindex:"""
        return type(self).from_file(path, key=key, single_vector=single_vector, transpose=transpose, id=self.id)

    @classmethod
    def _array_factory(cls, array, space=None, id=None):
        if type(array) is np.ndarray:
            pass
        elif issparse(array):
            array = array.toarray()
        else:
            array = np.array(array, ndmin=2)
        if array.ndim != 2:
            assert array.ndim == 1
            array = np.reshape(array, (1, -1))
        if space is None:
            return NumpyVectorArray(array, cls(array.shape[1], id))
        else:
            assert array.shape[1] == space.dim
            return NumpyVectorArray(array, space)

    @property
    def is_scalar(self):
        return self.dim == 1 and self.id is None


class NumpyVectorArrayView(NumpyVectorArray):

    is_view = True

    def __init__(self, array, ind):
        assert array.check_ind(ind)
        self.base = array
        self.ind = array.normalize_ind(ind)
        self.space = array.space

    def to_numpy(self, ensure_copy=False):
        result = self.base.to_numpy()[self.ind]
        if ensure_copy and not result.flags['OWNDATA']:
            result = result.copy()
        return result

    def __len__(self):
        return self.base.len_ind(self.ind)

    def __getitem__(self, ind):
        try:
            return self.base[self.base.sub_index(self.ind, ind)]
        except IndexError:
            raise IndexError('VectorArray index out of range')

    def __delitem__(self):
        raise ValueError('Cannot remove from NumpyVectorArrayView')

    def append(self, other, remove_from_other=False):
        raise ValueError('Cannot append to NumpyVectorArrayView')

    def copy(self, deep=False):
        return self.base.copy(_ind=self.ind, deep=deep)

    def scal(self, alpha):
        assert self.base.check_ind_unique(self.ind)
        self.base.scal(alpha, _ind=self.ind)

    def axpy(self, alpha, x):
        assert self.base.check_ind_unique(self.ind)
        self.base.axpy(alpha, x, _ind=self.ind)

    def inner(self, other, product=None):
        if product is not None:
            return product.apply2(self, other)
        return self.base.inner(other, _ind=self.ind)

    def pairwise_inner(self, other, product=None):
        if product is not None:
            return product.pairwise_apply2(self, other)
        return self.base.pairwise_inner(other, _ind=self.ind)

    def lincomb(self, coefficients):
        return self.base.lincomb(coefficients, _ind=self.ind)

    def _norm(self):
        return self.base._norm(_ind=self.ind)

    def _norm2(self):
        return self.base._norm2(_ind=self.ind)

    def sup_norm(self):
        return self.base.sup_norm(_ind=self.ind)

    def dofs(self, dof_indices):
        return self.base.dofs(dof_indices, _ind=self.ind)

    def amax(self):
        return self.base.amax(_ind=self.ind)

    def __str__(self):
        return self.base._array[:self.base._len][self.ind].__str__()

    def __add__(self, other):
        if isinstance(other, Number):
            assert other == 0
            return self.copy()
        assert self.dim == other.dim
        return NumpyVectorArray(self.base._array[self.ind]
                                + (other.base._array[other.ind] if other.is_view else other._array[:other._len]),
                                self.space)

    def __iadd__(self, other):
        assert self.dim == other.dim
        assert self.base.check_ind_unique(self.ind)
        if self.base._refcount[0] > 1:
            self._deep_copy()
        other_dtype = other.base._array.dtype if other.is_view else other._array.dtype
        common_dtype = np.promote_types(self.base._array.dtype, other_dtype)
        if self.base._array.dtype != common_dtype:
            self.base._array = self.base._array.astype(common_dtype)
        self.base.array[self.ind] += other.base._array[other.ind] if other.is_view else other._array[:other._len]
        return self

    __radd__ = __add__

    def __sub__(self, other):
        assert self.dim == other.dim
        return NumpyVectorArray(self.base._array[self.ind]
                                - (other.base._array[other.ind] if other.is_view else other._array[:other._len]),
                                self.space)

    def __isub__(self, other):
        assert self.dim == other.dim
        assert self.base.check_ind_unique(self.ind)
        if self.base._refcount[0] > 1:
            self._deep_copy()
        other_dtype = other.base._array.dtype if other.is_view else other._array.dtype
        common_dtype = np.promote_types(self.base._array.dtype, other_dtype)
        if self.base._array.dtype != common_dtype:
            self.base._array = self.base._array.astype(common_dtype)
        self.base._array[self.ind] -= other.base._array[other.ind] if other.is_view else other._array[:other._len]
        return self

    def __mul__(self, other):
        assert isinstance(other, Number) \
            or isinstance(other, np.ndarray) and other.shape == (len(self),)
        if type(other) is np.ndarray:
            other = other[:, np.newaxis]
        return NumpyVectorArray(self.base._array[self.ind] * other, self.space)

    def __imul__(self, other):
        assert isinstance(other, Number) \
            or isinstance(other, np.ndarray) and other.shape == (len(self),)
        if type(other) is np.ndarray:
            other = other[:, np.newaxis]
        assert self.base.check_ind_unique(self.ind)
        if self.base._refcount[0] > 1:
            self._deep_copy()
        other_dtype = other.dtype if isinstance(other, np.ndarray) else type(other)
        common_dtype = np.promote_types(self.base._array.dtype, other_dtype)
        if self.base._array.dtype != common_dtype:
            self.base._array = self.base._array.astype(common_dtype)
        self.base._array[self.ind] *= other
        return self

    def __neg__(self):
        return NumpyVectorArray(-self.base._array[self.ind], self.space)

    def __del__(self):
        return

    def __repr__(self):
        return f'NumpyVectorArrayView({self.to_numpy()}, {self.space})'



# This file is part of the pyMOR project (https://www.pymor.org).
# Copyright 2013-2021 pyMOR developers and contributors. All rights reserved.
# License: BSD 2-Clause License (https://opensource.org/licenses/BSD-2-Clause)

from itertools import product

import numpy as np
import pytest
from hypothesis import given, settings

from pymor.core.exceptions import QtMissing
from pymortests.base import might_exceed_deadline
from pymortests.fixtures.grid import hy_grids_with_visualize, hy_grid, hy_grid_and_dim_range_product, \
    hy_grid_and_dim_range_product_and_s_max_en, hy_grid_and_dim_range_product_and_s, \
    hy_grid_and_dim_range_product_and_s_to_e
from pymortests.pickling import assert_picklable_without_dumps_function


# monkey np.testing.assert_allclose to behave the same as np.allclose
# for some reason, the default atol of np.testing.assert_allclose is 0
# while it is 1e-8 for np.allclose

real_assert_allclose = np.testing.assert_allclose


def monkey_allclose(a, b, rtol=1.e-5, atol=1.e-8):
    real_assert_allclose(a, b, rtol=rtol, atol=atol)


np.testing.assert_allclose = monkey_allclose


@given(hy_grid)
def test_dim(grid):
    g = grid
    assert isinstance(g.dim, int)
    assert g.dim >= 0


@given(hy_grid)
def test_size(grid):
    g = grid
    for d in range(g.dim + 1):
        assert g.size(d) >= 0
    with pytest.raises(AssertionError):
        g.size(-1)
    with pytest.raises(AssertionError):
        g.size(g.dim + 1)


@given(hy_grid_and_dim_range_product())
def test_subentities_wrong_arguments(grid_and_dims):
    g, e, _ = grid_and_dims
    with pytest.raises(AssertionError):
        g.subentities(e, g.dim + 1)
    with pytest.raises(AssertionError):
        g.subentities(e, e - 1)
    with pytest.raises(AssertionError):
        g.subentities(g.dim + 1, 0)
    with pytest.raises(AssertionError):
        g.subentities(-1, 0)


@given(hy_grid_and_dim_range_product_and_s())
def test_subentities_shape(grid_and_dims):
    g, e, _, s = grid_and_dims
    assert g.subentities(e, s).ndim == 2
    assert g.subentities(e, s).shape[0] == g.size(e)


@given(hy_grid_and_dim_range_product_and_s())
def test_subentities_dtype(grid_and_dims):
    g, e, _, s = grid_and_dims
    assert g.subentities(e, s).dtype == np.dtype('int32')


@given(hy_grid_and_dim_range_product_and_s())
def test_subentities_entry_value_range(grid_and_dims):
    g, e, _, s = grid_and_dims
    np.testing.assert_array_less(g.subentities(e, s), g.size(s))
    np.testing.assert_array_less(-2, g.subentities(e, s))


@given(hy_grid_and_dim_range_product_and_s())
def test_subentities_entry_values_unique(grid_and_dims):
    g, e, _, s = grid_and_dims
    for S in g.subentities(e, s):
        S = S[S >= 0]
        assert S.size == np.unique(S).size


@given(hy_grid_and_dim_range_product())
def test_subentities_codim_d_codim_d(grid_and_dims):
    g, e, _ = grid_and_dims
    assert g.subentities(e, e).shape == (g.size(e), 1)
    np.testing.assert_array_equal(g.subentities(e, e).ravel(), np.arange(g.size(e)))


@given(hy_grid_and_dim_range_product())
def test_subentities_transitivity(grid_and_dims):
    g, e, _ = grid_and_dims
    for s in range(e + 1, g.dim + 1):
        for ss in range(s + 1, g.dim + 1):
            SE = g.subentities(e, s)
            SSE = g.subentities(e, ss)
            SESE = g.subentities(s, ss)
            for i in range(SE.shape[0]):
                for j in range(SE.shape[1]):
                    if SE[i, j] != -1:
                        assert set(SESE[SE[i, j]]) <= set(SSE[i]).union((-1,))


@given(hy_grid_and_dim_range_product())
def test_superentities_wrong_arguments(grid_and_dims):
    g, e, _ = grid_and_dims
    with pytest.raises(AssertionError):
        g.superentities(e, -1)
    with pytest.raises(AssertionError):
        g.superentities(e, e + 1)
    with pytest.raises(AssertionError):
        g.superentities(g.dim + 1, 0)
    with pytest.raises(AssertionError):
        g.superentities(-1, 0)
    with pytest.raises(AssertionError):
        g.superentities(-1, -2)


@given(hy_grid_and_dim_range_product_and_s_to_e())
def test_superentities_shape(grid_and_dims):
    g, e, _, s = grid_and_dims
    assert g.superentities(e, s).ndim == 2
    assert g.superentities(e, s).shape[0] == g.size(e)
    assert g.superentities(e, s).shape[1] > 0
    assert g.superentities(1, 0).shape[1] <= 2


@given(hy_grid_and_dim_range_product_and_s_to_e())
def test_superentities_dtype(grid_and_dims):
    g, e, _, s = grid_and_dims
    assert g.superentities(e, s).dtype == np.dtype('int32')


@given(hy_grid_and_dim_range_product_and_s_to_e())
def test_superentities_entry_value_range(grid_and_dims):
    g, e, _, s = grid_and_dims
    np.testing.assert_array_less(g.superentities(e, s), g.size(s))
    np.testing.assert_array_less(-2, g.superentities(e, s))


@given(hy_grid_and_dim_range_product_and_s_to_e())
def test_superentities_entry_values_unique(grid_and_dims):
    g, e, _, s = grid_and_dims
    for S in g.superentities(e, s):
        S = S[S >= 0]
        assert S.size == np.unique(S).size


@given(hy_grid_and_dim_range_product_and_s_to_e())
def test_superentities_entries_sorted(grid_and_dims):
    g, e, _, s = grid_and_dims
    for S in g.superentities(e, s):
        i = 0
        while (i + 1 < len(S)) and (S[i] < S[i + 1]):
            i += 1
        assert (i + 1 == len(S)) or (S[i + 1] == -1)
        if i + 1 < len(S):
            np.testing.assert_array_equal(S[i + 1:], -1)


@given(hy_grid_and_dim_range_product())
def test_superentities_codim_d_codim_d(grid_and_dims):
    g, e, _ = grid_and_dims
    assert g.superentities(e, e).shape == (g.size(e), 1)
    np.testing.assert_array_equal(g.superentities(e, e).ravel(), np.arange(g.size(e)))


@given(hy_grid_and_dim_range_product_and_s_to_e())
def test_superentities_each_entry_superentity(grid_and_dims):
    g, e, _, s = grid_and_dims
    SE = g.superentities(e, s)
    SUBE = g.subentities(s, e)
    for i in range(SE.shape[0]):
        for se in SE[i]:
            if se != -1:
                assert i in SUBE[se]


@given(hy_grid_and_dim_range_product_and_s_to_e())
def test_superentities_each_superentity_has_entry(grid_and_dims):
    g, e, _, s = grid_and_dims
    SE = g.superentities(e, s)
    SUBE = g.subentities(s, e)
    for i in range(SUBE.shape[0]):
        for se in SUBE[i]:
            if se != -1:
                assert i in SE[se]


@given(hy_grid_and_dim_range_product())
def test_superentity_indices_wrong_arguments(grid_and_dims):
    g, e, _ = grid_and_dims
    with pytest.raises(AssertionError):
        g.superentity_indices(e, -1)
    with pytest.raises(AssertionError):
        g.superentity_indices(e, e + 1)
    with pytest.raises(AssertionError):
        g.superentity_indices(g.dim + 1, 0)
    with pytest.raises(AssertionError):
        g.superentity_indices(-1, 0)
    with pytest.raises(AssertionError):
        g.superentity_indices(-1, -2)


@given(hy_grid_and_dim_range_product_and_s_to_e())
def test_superentity_indices_shape(grid_and_dims):
    g, e, _, s = grid_and_dims
    assert g.superentity_indices(e, s).shape == g.superentities(e, s).shape


@given(hy_grid_and_dim_range_product_and_s_to_e())
def test_superentity_indices_dtype(grid_and_dims):
    g, e, _, s = grid_and_dims
    assert g.superentity_indices(e, s).dtype == np.dtype('int32')


@given(hy_grid_and_dim_range_product_and_s_to_e())
def test_superentity_indices_valid_entries(grid_and_dims):
    g, e, _, s = grid_and_dims
    SE = g.superentities(e, s)
    SEI = g.superentity_indices(e, s)
    SUBE = g.subentities(s, e)
    for index, superentity in np.ndenumerate(SE):
        if superentity > -1:
            assert SUBE[superentity, SEI[index]] == index[0]


@given(hy_grid_and_dim_range_product())
def test_neighbours_wrong_arguments(grid_and_dims):
    g, e, n = grid_and_dims
    with pytest.raises(AssertionError):
        g.neighbours(e, n, -1)
    with pytest.raises(AssertionError):
        g.neighbours(e, n, g.dim + 1)
    with pytest.raises(AssertionError):
        g.neighbours(e, n, e - 1)
    with pytest.raises(AssertionError):
        g.neighbours(e, n, n - 1)
    with pytest.raises(AssertionError):
        g.neighbours(e, g.dim + 1, g.dim)
    with pytest.raises(AssertionError):
        g.neighbours(e, -1, g.dim)
    with pytest.raises(AssertionError):
        g.neighbours(g.dim + 1, g.dim, g.dim)
    with pytest.raises(AssertionError):
        g.neighbours(-1, 0, g.dim)


@might_exceed_deadline(2000)
@given(hy_grid_and_dim_range_product_and_s_max_en())
def test_neighbours_shape(grid_and_dims):
    g, e, n, s = grid_and_dims
    assert g.neighbours(e, n, s).ndim == 2
    assert g.neighbours(e, n, s).shape[0] == g.size(e)


@might_exceed_deadline(2000)
@given(hy_grid_and_dim_range_product_and_s_max_en())
def test_neighbours_dtype(grid_and_dims):
    g, e, n, s = grid_and_dims
    assert g.neighbours(e, n, s).dtype == np.dtype('int32')


@might_exceed_deadline(2000)
@given(hy_grid_and_dim_range_product_and_s_max_en())
def test_neighbours_entry_value_range(grid_and_dims):
    g, e, n, s = grid_and_dims
    np.testing.assert_array_less(g.neighbours(e, n, s), g.size(n))
    np.testing.assert_array_less(-2, g.neighbours(e, n, s))


@might_exceed_deadline(2000)
@given(hy_grid_and_dim_range_product_and_s_max_en())
def test_neighbours_entry_values_unique(grid_and_dims):
    g, e, n, s = grid_and_dims
    for S in g.neighbours(e, n, s):
        S = S[S >= 0]
        assert S.size == np.unique(S).size


@might_exceed_deadline(5000)
@given(hy_grid_and_dim_range_product_and_s_max_en())
def test_neighbours_each_entry_neighbour(grid_and_dims):
    g, e, n, s = grid_and_dims
    N = g.neighbours(e, n, s)
    ESE = g.subentities(e, s)
    NSE = g.subentities(n, s)
    for index, neigh in np.ndenumerate(N):
        if neigh > -1:
            inter = set(ESE[index[0]]).intersection(set(NSE[neigh]))
            if -1 in inter:
                assert len(inter) > 1
            else:
                assert len(inter) > 0


@might_exceed_deadline(4000)
@given(hy_grid_and_dim_range_product_and_s_max_en())
def test_neighbours_each_neighbour_has_entry(grid_and_dims):
    g, e, n, s = grid_and_dims
    N = g.neighbours(e, n, s)
    SUE = g.superentities(s, e)
    SUN = g.superentities(s, n)
    if e != n:
        for si in range(SUE.shape[0]):
            for ei, ni in product(SUE[si], SUN[si]):
                if ei != -1 and ni != -1:
                    assert ni in N[ei],\
                        f'Failed for\n{g}\ne={e}, n={n}, s={s}, ei={ei}, ni={ni}'
    else:
        for si in range(SUE.shape[0]):
            for ei, ni in product(SUE[si], SUN[si]):
                if ei != ni and ei != -1 and ni != -1:
                    assert ni in N[ei],\
                        f'Failed for\n{g}\ne={e}, n={n}, s={s}, ei={ei}, ni={ni}'


@might_exceed_deadline(2000)
@given(hy_grid_and_dim_range_product_and_s_max_en())
def test_neighbours_not_neighbour_of_itself(grid_and_dims):
    g, e, _, s = grid_and_dims
    N = g.neighbours(e, e, s)
    for ei, E in enumerate(N):
        assert ei not in E,\
            f'Failed for\n{g}\ne={e}, s={s}, ei={ei}, E={E}'


@given(hy_grid)
def test_boundary_mask_wrong_arguments(grid):
    g = grid
    with pytest.raises(AssertionError):
        g.boundary_mask(-1)
    with pytest.raises(AssertionError):
        g.boundary_mask(g.dim + 1)


@given(hy_grid)
def test_boundary_mask_shape(grid):
    g = grid
    for d in range(g.dim + 1):
        assert g.boundary_mask(d).shape == (g.size(d),)


@given(hy_grid)
def test_boundary_mask_dtype(grid):
    g = grid
    for d in range(g.dim + 1):
        g.boundary_mask(d).dtype == np.dtype('bool')


@given(hy_grid)
def test_boundary_mask_entries_codim1(grid):
    g = grid
    BM = g.boundary_mask(1)
    SE = g.superentities(1, 0)
    for ei, b in enumerate(BM):
        E = SE[ei]
        assert (E[E > -1].size <= 1) == b


@given(hy_grid)
def test_boundary_mask_entries_codim0(grid):
    g = grid
    BM0 = g.boundary_mask(0)
    BM1 = g.boundary_mask(1)
    SE = g.subentities(0, 1)
    for ei, b in enumerate(BM0):
        S = SE[ei]
        assert np.any(BM1[S[S > -1]]) == b


@given(hy_grid)
def test_boundary_mask_entries_codim_d(grid):
    g = grid
    for d in range(2, g.dim + 1):
        BMD = g.boundary_mask(d)
        BM1 = g.boundary_mask(1)
        SE = g.superentities(d, 1)
        for ei, b in enumerate(BMD):
            S = SE[ei]
            assert np.any(BM1[S[S > -1]]) == b


@given(hy_grid)
def test_boundaries_wrong_arguments(grid):
    g = grid
    with pytest.raises(AssertionError):
        g.boundaries(-1)
    with pytest.raises(AssertionError):
        g.boundaries(g.dim + 1)


@given(hy_grid)
def test_boundaries_shape(grid):
    g = grid
    for d in range(g.dim + 1):
        assert len(g.boundaries(d).shape) == 1


@given(hy_grid)
def test_boundaries_dtype(grid):
    g = grid
    for d in range(g.dim + 1):
        assert g.boundaries(d).dtype == np.dtype('int32')


@might_exceed_deadline(2000)
@given(hy_grid)
def test_boundaries_entry_value_range(grid):
    g = grid
    for d in range(g.dim + 1):
        np.testing.assert_array_less(g.boundaries(d), g.size(d))
        np.testing.assert_array_less(-1, g.boundaries(d))


@given(hy_grid)
def test_boundaries_entries(grid):
    g = grid
    for d in range(g.dim + 1):
        np.testing.assert_array_equal(np.where(g.boundary_mask(d))[0], g.boundaries(d))


@given(hy_grid)
def test_pickle(grid):
    assert_picklable_without_dumps_function(grid)


@settings(deadline=None)
@given(hy_grids_with_visualize)
def test_visualize(grids_with_visualize):
    import sys
    sys._called_from_test = True

    def nop(*args, **kwargs):
        pass
    try:
        from matplotlib import pyplot
        if sys.version_info[:2] > (3, 7) or (
                sys.version_info[0] == 3 and sys.version_info[1] == 6):
            pyplot.ion()
        else:
            # the ion switch results in interpreter segfaults during multiple
            # demo tests on 3.7 -> fall back on old monkeying solution
            pyplot.show = nop
    except ImportError:
        pass

    try:
        g = grids_with_visualize
        U = np.ones(g.size(g.dim))
        g.visualize(U, g.dim)
    except QtMissing:
        pytest.xfail("Qt missing")



# This file is part of the pyMOR project (https://www.pymor.org).
# Copyright 2013-2021 pyMOR developers and contributors. All rights reserved.
# License: BSD 2-Clause License (https://opensource.org/licenses/BSD-2-Clause)

from pymor.core.config import config
from pymor.tools.deprecated import Deprecated

if config.HAVE_PYAMG:

    import numpy as np
    import pyamg

    from pymor.algorithms.genericsolvers import _parse_options
    from pymor.core.defaults import defaults
    from pymor.core.exceptions import InversionError
    from pymor.operators.numpy import NumpyMatrixOperator

    @defaults('tol', 'maxiter', 'verb', 'rs_strength', 'rs_CF',
              'rs_postsmoother', 'rs_max_levels', 'rs_max_coarse', 'rs_coarse_solver',
              'rs_cycle', 'rs_accel', 'rs_tol', 'rs_maxiter',
              'sa_symmetry', 'sa_strength', 'sa_aggregate', 'sa_smooth',
              'sa_presmoother', 'sa_postsmoother', 'sa_improve_candidates', 'sa_max_levels',
              'sa_max_coarse', 'sa_diagonal_dominance', 'sa_coarse_solver', 'sa_cycle',
              'sa_accel', 'sa_tol', 'sa_maxiter')
    def solver_options(tol=1e-5,
                       maxiter=400,
                       verb=False,
                       rs_strength=('classical', {'theta': 0.25}),
                       rs_CF='RS',
                       rs_presmoother=('gauss_seidel', {'sweep': 'symmetric'}),
                       rs_postsmoother=('gauss_seidel', {'sweep': 'symmetric'}),
                       rs_max_levels=10,
                       rs_max_coarse=500,
                       rs_coarse_solver='pinv2',
                       rs_cycle='V',
                       rs_accel=None,
                       rs_tol=1e-5,
                       rs_maxiter=100,
                       sa_symmetry='hermitian',
                       sa_strength='symmetric',
                       sa_aggregate='standard',
                       sa_smooth=('jacobi', {'omega': 4.0/3.0}),
                       sa_presmoother=('block_gauss_seidel', {'sweep': 'symmetric'}),
                       sa_postsmoother=('block_gauss_seidel', {'sweep': 'symmetric'}),
                       sa_improve_candidates=(('block_gauss_seidel', {'sweep': 'symmetric', 'iterations': 4}), None),
                       sa_max_levels=10,
                       sa_max_coarse=500,
                       sa_diagonal_dominance=False,
                       sa_coarse_solver='pinv2',
                       sa_cycle='V',
                       sa_accel=None,
                       sa_tol=1e-5,
                       sa_maxiter=100):
        """Returns available solvers with default |solver_options| for the PyAMG backend.

        Parameters
        ----------
        tol
            Tolerance for `PyAMG <http://pyamg.github.io/>`_ blackbox solver.
        maxiter
            Maximum iterations for `PyAMG <http://pyamg.github.io/>`_ blackbox solver.
        verb
            Verbosity flag for `PyAMG <http://pyamg.github.io/>`_ blackbox solver.
        rs_strength
            Parameter for `PyAMG <http://pyamg.github.io/>`_ Ruge-Stuben solver.
        rs_CF
            Parameter for `PyAMG <http://pyamg.github.io/>`_ Ruge-Stuben solver.
        rs_presmoother
            Parameter for `PyAMG <http://pyamg.github.io/>`_ Ruge-Stuben solver.
        rs_postsmoother
            Parameter for `PyAMG <http://pyamg.github.io/>`_ Ruge-Stuben solver.
        rs_max_levels
            Parameter for `PyAMG <http://pyamg.github.io/>`_ Ruge-Stuben solver.
        rs_max_coarse
            Parameter for `PyAMG <http://pyamg.github.io/>`_ Ruge-Stuben solver.
        rs_coarse_solver
            Parameter for `PyAMG <http://pyamg.github.io/>`_ Ruge-Stuben solver.
        rs_cycle
            Parameter for `PyAMG <http://pyamg.github.io/>`_ Ruge-Stuben solver.
        rs_accel
            Parameter for `PyAMG <http://pyamg.github.io/>`_ Ruge-Stuben solver.
        rs_tol
            Parameter for `PyAMG <http://pyamg.github.io/>`_ Ruge-Stuben solver.
        rs_maxiter
            Parameter for `PyAMG <http://pyamg.github.io/>`_ Ruge-Stuben solver.
        sa_symmetry
            Parameter for `PyAMG <http://pyamg.github.io/>`_ Smoothed-Aggregation solver.
        sa_strength
            Parameter for `PyAMG <http://pyamg.github.io/>`_ Smoothed-Aggregation solver.
        sa_aggregate
            Parameter for `PyAMG <http://pyamg.github.io/>`_ Smoothed-Aggregation solver.
        sa_smooth
            Parameter for `PyAMG <http://pyamg.github.io/>`_ Smoothed-Aggregation solver.
        sa_presmoother
            Parameter for `PyAMG <http://pyamg.github.io/>`_ Smoothed-Aggregation solver.
        sa_postsmoother
            Parameter for `PyAMG <http://pyamg.github.io/>`_ Smoothed-Aggregation solver.
        sa_improve_candidates
            Parameter for `PyAMG <http://pyamg.github.io/>`_ Smoothed-Aggregation solver.
        sa_max_levels
            Parameter for `PyAMG <http://pyamg.github.io/>`_ Smoothed-Aggregation solver.
        sa_max_coarse
            Parameter for `PyAMG <http://pyamg.github.io/>`_ Smoothed-Aggregation solver.
        sa_diagonal_dominance
            Parameter for `PyAMG <http://pyamg.github.io/>`_ Smoothed-Aggregation solver.
        sa_coarse_solver
            Parameter for `PyAMG <http://pyamg.github.io/>`_ Smoothed-Aggregation solver.
        sa_cycle
            Parameter for `PyAMG <http://pyamg.github.io/>`_ Smoothed-Aggregation solver.
        sa_accel
            Parameter for `PyAMG <http://pyamg.github.io/>`_ Smoothed-Aggregation solver.
        sa_tol
            Parameter for `PyAMG <http://pyamg.github.io/>`_ Smoothed-Aggregation solver.
        sa_maxiter
            Parameter for `PyAMG <http://pyamg.github.io/>`_ Smoothed-Aggregation solver.

        Returns
        -------
        A dict of available solvers with default |solver_options|.
        """
        return {'pyamg_solve': {'type': 'pyamg_solve',
                                'tol': tol,
                                'maxiter': maxiter},
                'pyamg_rs':    {'type': 'pyamg_rs',
                                'strength': rs_strength,
                                'CF': rs_CF,
                                'presmoother': rs_presmoother,
                                'postsmoother': rs_postsmoother,
                                'max_levels': rs_max_levels,
                                'max_coarse': rs_max_coarse,
                                'coarse_solver': rs_coarse_solver,
                                'cycle': rs_cycle,
                                'accel': rs_accel,
                                'tol': rs_tol,
                                'maxiter': rs_maxiter},
                'pyamg_sa':    {'type': 'pyamg_sa',
                                'symmetry': sa_symmetry,
                                'strength': sa_strength,
                                'aggregate': sa_aggregate,
                                'smooth': sa_smooth,
                                'presmoother': sa_presmoother,
                                'postsmoother': sa_postsmoother,
                                'improve_candidates': sa_improve_candidates,
                                'max_levels': sa_max_levels,
                                'max_coarse': sa_max_coarse,
                                'diagonal_dominance': sa_diagonal_dominance,
                                'coarse_solver': sa_coarse_solver,
                                'cycle': sa_cycle,
                                'accel': sa_accel,
                                'tol': sa_tol,
                                'maxiter': sa_maxiter}}

    @Deprecated('pyamg bindings will be removed after the 2021.1 release')
    @defaults('check_finite', 'default_solver')
    def apply_inverse(op, V, initial_guess=None, options=None, least_squares=False,
                      check_finite=True, default_solver='pyamg_solve'):
        """Solve linear equation system.

        Applies the inverse of `op` to the vectors in `V` using PyAMG.

        Parameters
        ----------
        op
            The linear, non-parametric |Operator| to invert.
        V
            |VectorArray| of right-hand sides for the equation system.
        initial_guess
            |VectorArray| with the same length as `V` containing initial guesses
            for the solution.  Some implementations of `apply_inverse` may
            ignore this parameter.  If `None` a solver-dependent default is used.
        options
            The |solver_options| to use (see :func:`solver_options`).
        least_squares
            Must be `False`.
        check_finite
            Test if solution only contains finite values.
        default_solver
            Default solver to use (pyamg_solve, pyamg_rs, pyamg_sa).

        Returns
        -------
        |VectorArray| of the solution vectors.
        """
        assert V in op.range
        assert initial_guess is None or initial_guess in op.source and len(initial_guess) == len(V)

        if least_squares:
            raise NotImplementedError

        if isinstance(op, NumpyMatrixOperator):
            matrix = op.matrix
        else:
            from pymor.algorithms.to_matrix import to_matrix
            matrix = to_matrix(op)

        options = _parse_options(options, solver_options(), default_solver, None, least_squares)

        V = V.to_numpy()
        promoted_type = np.promote_types(matrix.dtype, V.dtype)
        R = np.empty((len(V), matrix.shape[1]), dtype=promoted_type)

        if options['type'] == 'pyamg_solve':
            if len(V) > 0:
                V_iter = iter(enumerate(V))
                R[0], ml = pyamg.solve(matrix, next(V_iter)[1],
                                       tol=options['tol'],
                                       maxiter=options['maxiter'],
                                       return_solver=True)
                for i, VV in V_iter:
                    R[i] = pyamg.solve(matrix, VV,
                                       tol=options['tol'],
                                       maxiter=options['maxiter'],
                                       existing_solver=ml)
        elif options['type'] == 'pyamg_rs':
            ml = pyamg.ruge_stuben_solver(matrix,
                                          strength=options['strength'],
                                          CF=options['CF'],
                                          presmoother=options['presmoother'],
                                          postsmoother=options['postsmoother'],
                                          max_levels=options['max_levels'],
                                          max_coarse=options['max_coarse'],
                                          coarse_solver=options['coarse_solver'])
            for i, VV in enumerate(V):
                R[i] = ml.solve(VV,
                                tol=options['tol'],
                                maxiter=options['maxiter'],
                                cycle=options['cycle'],
                                accel=options['accel'])
        elif options['type'] == 'pyamg_sa':
            ml = pyamg.smoothed_aggregation_solver(matrix,
                                                   symmetry=options['symmetry'],
                                                   strength=options['strength'],
                                                   aggregate=options['aggregate'],
                                                   smooth=options['smooth'],
                                                   presmoother=options['presmoother'],
                                                   postsmoother=options['postsmoother'],
                                                   improve_candidates=options['improve_candidates'],
                                                   max_levels=options['max_levels'],
                                                   max_coarse=options['max_coarse'],
                                                   diagonal_dominance=options['diagonal_dominance'])
            for i, VV in enumerate(V):
                R[i] = ml.solve(VV,
                                tol=options['tol'],
                                maxiter=options['maxiter'],
                                cycle=options['cycle'],
                                accel=options['accel'])
        else:
            raise ValueError('Unknown solver type')

        if check_finite:
            if not np.isfinite(np.sum(R)):
                raise InversionError('Result contains non-finite values')

        return op.source.from_numpy(R)



# This file is part of the pyMOR project (https://www.pymor.org).
# Copyright 2013-2021 pyMOR developers and contributors. All rights reserved.
# License: BSD 2-Clause License (https://opensource.org/licenses/BSD-2-Clause)
from pathlib import Path

from pymor.core.config import config
from pymor.core.defaults import defaults
from pymor.tools.io import change_to_directory

if config.HAVE_NGSOLVE:
    import ngsolve as ngs
    import numpy as np

    from pymor.core.base import ImmutableObject
    from pymor.operators.list import LinearComplexifiedListVectorArrayOperatorBase
    from pymor.vectorarrays.interface import VectorArray
    from pymor.vectorarrays.numpy import NumpyVectorSpace
    from pymor.vectorarrays.list import CopyOnWriteVector, ComplexifiedVector, ComplexifiedListVectorSpace

    class NGSolveVectorCommon:
        def amax(self):
            A = np.abs(self.to_numpy())
            max_ind = np.argmax(A)
            max_val = A[max_ind]
            return max_ind, max_val

        def dofs(self, dof_indices):
            return self.to_numpy()[dof_indices]

    class NGSolveVector(NGSolveVectorCommon, CopyOnWriteVector):
        """Wraps a NGSolve BaseVector to make it usable with ListVectorArray."""

        def __init__(self, impl):
            self.impl = impl

        @classmethod
        def from_instance(cls, instance):
            return cls(instance.impl)

        def _copy_data(self):
            new_impl = ngs.GridFunction(self.impl.space)
            new_impl.vec.data = self.impl.vec
            self.impl = new_impl

        def to_numpy(self, ensure_copy=False):
            if ensure_copy:
                return self.impl.vec.FV().NumPy().copy()
            self._copy_data_if_needed()
            return self.impl.vec.FV().NumPy()

        def _scal(self, alpha):
            self.impl.vec.data = float(alpha) * self.impl.vec

        def _axpy(self, alpha, x):
            self.impl.vec.data = self.impl.vec + float(alpha) * x.impl.vec

        def inner(self, other):
            return self.impl.vec.InnerProduct(other.impl.vec)

        def norm(self):
            return self.impl.vec.Norm()

        def norm2(self):
            return self.impl.vec.Norm() ** 2

    class ComplexifiedNGSolveVector(NGSolveVectorCommon, ComplexifiedVector):
        pass

    class NGSolveVectorSpace(ComplexifiedListVectorSpace):

        real_vector_type = NGSolveVector
        vector_type = ComplexifiedNGSolveVector

        def __init__(self, V, id='STATE'):
            self.__auto_init(locals())

        def __eq__(self, other):
            return type(other) is NGSolveVectorSpace and self.V == other.V and self.id == other.id

        def __hash__(self):
            return hash(self.V) + hash(self.id)

        @property
        def value_dim(self):
            u = self.V.TrialFunction()
            if isinstance(u, list):
                return u[0].dim
            else:
                return u.dim

        @property
        def dim(self):
            return self.V.ndofglobal * self.value_dim

        @classmethod
        def space_from_vector_obj(cls, vec, id):
            return cls(vec.space, id)

        def real_zero_vector(self):
            impl = ngs.GridFunction(self.V)
            return NGSolveVector(impl)

        def real_make_vector(self, obj):
            return NGSolveVector(obj)

        def real_vector_from_numpy(self, data, ensure_copy=False):
            v = self.real_zero_vector()
            v.to_numpy()[:] = data
            return v

    class NGSolveMatrixOperator(LinearComplexifiedListVectorArrayOperatorBase):
        """Wraps a NGSolve matrix as an |Operator|."""

        def __init__(self, matrix, range, source, solver_options=None, name=None):
            self.__auto_init(locals())

        @defaults('default_solver')
        def _prepare_apply(self, U, mu, kind, least_squares=False, default_solver=''):
            if kind == 'apply_inverse':
                if least_squares:
                    raise NotImplementedError
                solver = self.solver_options.get('inverse', default_solver) if self.solver_options else default_solver
                inv = self.matrix.Inverse(self.source.V.FreeDofs(), inverse=solver)
                return inv

        def _real_apply_one_vector(self, u, mu=None, prepare_data=None):
            r = self.range.real_zero_vector()
            self.matrix.Mult(u.impl.vec, r.impl.vec)
            return r

        def _real_apply_adjoint_one_vector(self, v, mu=None, prepare_data=None):
            u = self.source.real_zero_vector()
            try:
                mat = self.matrix.Transpose()
            except AttributeError:
                mat = self.matrix.T
            mat.Mult(v.impl.vec, u.impl.vec)
            return u

        def _real_apply_inverse_one_vector(self, v, mu=None, initial_guess=None,
                                           least_squares=False, prepare_data=None):
            inv = prepare_data
            r = self.source.real_zero_vector()
            r.impl.vec.data = inv * v.impl.vec
            return r

        def _assemble_lincomb(self, operators, coefficients, identity_shift=0., solver_options=None, name=None):
            if not all(isinstance(op, NGSolveMatrixOperator) for op in operators):
                return None
            if identity_shift != 0:
                return None

            matrix = operators[0].matrix.CreateMatrix()
            matrix.AsVector().data = float(coefficients[0]) * matrix.AsVector()
            for op, c in zip(operators[1:], coefficients[1:]):
                matrix.AsVector().data += float(c) * op.matrix.AsVector()
            return NGSolveMatrixOperator(matrix, self.range, self.source, solver_options=solver_options, name=name)

        def as_vector(self, copy=True):
            vec = self.matrix.AsVector().FV().NumPy()
            return NumpyVectorSpace.make_array(vec.copy() if copy else vec)

    class NGSolveVisualizer(ImmutableObject):
        """Visualize an NGSolve grid function."""

        def __init__(self, mesh, fespace):
            self.__auto_init(locals())
            self.space = NGSolveVectorSpace(fespace)

        def visualize(self, U, legend=None, separate_colorbars=True, filename=None, block=True):
            """Visualize the provided data."""
            if isinstance(U, VectorArray):
                U = (U,)
            assert all(u in self.space for u in U)
            if any(len(u) != 1 for u in U):
                raise NotImplementedError
            if any(u._list[0].imag_part is not None for u in U):
                raise NotImplementedError
            if legend is None:
                legend = [f'VectorArray{i}' for i in range(len(U))]
            if isinstance(legend, str):
                legend = [legend]
            assert len(legend) == len(U)
            legend = [l.replace(' ', '_') for l in legend]  # NGSolve GUI will fail otherwise

            if filename:
                # ngsolve unconditionally appends ".vtk"
                filename = Path(filename).resolve()
                if filename.suffix == '.vtk':
                    filename = filename.parent / filename.stem
                else:
                    self.logger.warning(f'NGSolve set VTKOutput filename to {filename}.vtk')
                coeffs = [u._list[0].real_part.impl for u in U]
                # ngsolve cannot handle full paths for filenames
                with change_to_directory(filename.parent):
                    vtk = ngs.VTKOutput(ma=self.mesh, coefs=coeffs, names=legend, filename=str(filename), subdivision=0)
                    vtk.Do()
            else:
                if not separate_colorbars:
                    raise NotImplementedError

                for u, name in zip(U, legend):
                    ngs.Draw(u._list[0].real_part.impl, self.mesh, name=name)



# This file is part of the pyMOR project (https://www.pymor.org).
# Copyright 2013-2021 pyMOR developers and contributors. All rights reserved.
# License: BSD 2-Clause License (https://opensource.org/licenses/BSD-2-Clause)

"""Module for SVD method of operators represented by |VectorArrays|."""

import numpy as np
import scipy.linalg as spla

from pymor.algorithms.gram_schmidt import gram_schmidt
from pymor.core.defaults import defaults
from pymor.core.logger import getLogger
from pymor.operators.interface import Operator
from pymor.vectorarrays.interface import VectorArray


@defaults('rtol', 'atol', 'l2_err')
def method_of_snapshots(A, product=None, modes=None, rtol=1e-7, atol=0., l2_err=0.):
    """SVD of a |VectorArray| using the method of snapshots.

    Viewing the |VectorArray| `A` as a `A.dim` x `len(A)` matrix, the
    return value of this method is the singular value decomposition of
    `A`, where the inner product on R^(`dim(A)`) is given by `product`
    and the inner product on R^(`len(A)`) is the Euclidean inner
    product.

    .. warning::

        The left singular vectors may not be numerically orthonormal for
        ill-conditioned `A`.

    Parameters
    ----------
    A
        The |VectorArray| for which the SVD is to be computed.
    product
        Inner product |Operator| w.r.t. which the SVD is computed.
    modes
        If not `None`, at most the first `modes` singular values and
        vectors are returned.
    rtol
        Singular values smaller than this value multiplied by the
        largest singular value are ignored.
    atol
        Singular values smaller than this value are ignored.
    l2_err
        Do not return more modes than needed to bound the
        l2-approximation error by this value. I.e. the number of
        returned modes is at most ::

            argmin_N { sum_{n=N+1}^{infty} s_n^2 <= l2_err^2 }

        where `s_n` denotes the n-th singular value.

    Returns
    -------
    U
        |VectorArray| of left singular vectors.
    s
        One-dimensional |NumPy array| of singular values.
    Vh
        |NumPy array| of right singular vectors.
    """
    assert isinstance(A, VectorArray)
    assert product is None or isinstance(product, Operator)
    assert modes is None or modes <= len(A)

    if A.dim == 0 or len(A) == 0:
        return A.space.empty(), np.array([]), np.zeros((0, len(A)))

    logger = getLogger('pymor.algorithms.svd_va.method_of_snapshots')

    with logger.block(f'Computing Gramian ({len(A)} vectors) ...'):
        B = A.gramian(product)

    with logger.block('Computing eigenvalue decomposition ...'):
        eigvals = (None
                   if modes is None or l2_err > 0.
                   else (len(B) - modes, len(B) - 1))

        evals, V = spla.eigh(B, overwrite_a=True, turbo=True, eigvals=eigvals)
        evals = evals[::-1]
        V = V.T[::-1, :]

        tol = max(rtol ** 2 * evals[0], atol ** 2)
        above_tol = np.where(evals >= tol)[0]
        if len(above_tol) == 0:
            return A.space.empty(), np.array([]), np.zeros((0, len(A)))
        last_above_tol = above_tol[-1]

        errs = np.concatenate((np.cumsum(evals[::-1])[::-1], [0.]))
        below_err = np.where(errs <= l2_err**2)[0]
        first_below_err = below_err[0]

        selected_modes = min(first_below_err, last_above_tol + 1)
        if modes is not None:
            selected_modes = min(selected_modes, modes)

        if selected_modes > A.dim:
            logger.warning('Number of computed singular vectors larger than array dimension! Truncating ...')
            selected_modes = A.dim

        s = np.sqrt(evals[:selected_modes])
        V = V[:selected_modes]
        Vh = V.conj()

    with logger.block(f'Computing left-singular vectors ({len(V)} vectors) ...'):
        U = A.lincomb(V / s[:, np.newaxis])

    return U, s, Vh


@defaults('rtol', 'atol', 'l2_err')
def qr_svd(A, product=None, modes=None, rtol=4e-8, atol=0., l2_err=0.):
    """SVD of a |VectorArray| using Gram-Schmidt orthogonalization.

    Viewing the |VectorArray| `A` as a `A.dim` x `len(A)` matrix, the
    return value of this method is the singular value decomposition of
    `A`, where the inner product on R^(`dim(A)`) is given by `product`
    and the inner product on R^(`len(A)`) is the Euclidean inner
    product.

    Parameters
    ----------
    A
        The |VectorArray| for which the SVD is to be computed.
    product
        Inner product |Operator| w.r.t. which the left singular vectors
        are computed.
    modes
        If not `None`, at most the first `modes` singular values and
        vectors are returned.
    rtol
        Singular values smaller than this value multiplied by the
        largest singular value are ignored.
    atol
        Singular values smaller than this value are ignored.
    l2_err
        Do not return more modes than needed to bound the
        l2-approximation error by this value. I.e. the number of
        returned modes is at most ::

            argmin_N { sum_{n=N+1}^{infty} s_n^2 <= l2_err^2 }

        where `s_n` denotes the n-th singular value.

    Returns
    -------
    U
        |VectorArray| of left singular vectors.
    s
        One-dimensional |NumPy array| of singular values.
    Vh
        |NumPy array| of right singular vectors.
    """
    assert isinstance(A, VectorArray)
    assert product is None or isinstance(product, Operator)
    assert modes is None or modes <= len(A)

    if A.dim == 0 or len(A) == 0:
        return A.space.empty(), np.array([]), np.zeros((0, len(A)))

    logger = getLogger('pymor.algorithms.svd_va.qr_svd')

    with logger.block('Computing QR decomposition ...'):
        Q, R = gram_schmidt(A, product=product, return_R=True, check=False)

    with logger.block('Computing SVD of R ...'):
        U2, s, Vh = spla.svd(R, lapack_driver='gesvd')

    with logger.block('Choosing the number of modes ...'):
        tol = max(rtol * s[0], atol)
        above_tol = np.where(s >= tol)[0]
        if len(above_tol) == 0:
            return A.space.empty(), np.array([]), np.zeros((0, len(A)))
        last_above_tol = above_tol[-1]

        errs = np.concatenate((np.cumsum(s[::-1] ** 2)[::-1], [0.]))
        below_err = np.where(errs <= l2_err**2)[0]
        first_below_err = below_err[0]

        selected_modes = min(first_below_err, last_above_tol + 1)
        if modes is not None:
            selected_modes = min(selected_modes, modes)

        U2 = U2[:, :selected_modes]
        s = s[:selected_modes]
        Vh = Vh[:selected_modes]

    with logger.block(f'Computing left singular vectors ({selected_modes} modes) ...'):
        U = Q.lincomb(U2.T)

    return U, s, Vh



# This file is part of the pyMOR project (https://www.pymor.org).
# Copyright 2013-2021 pyMOR developers and contributors. All rights reserved.
# License: BSD 2-Clause License (https://opensource.org/licenses/BSD-2-Clause)

from pymor.discretizers.builtin.cg import discretize_stationary_cg, discretize_instationary_cg
from pymor.discretizers.builtin.fv import discretize_stationary_fv, discretize_instationary_fv
from pymor.discretizers.builtin.grids.gmsh import load_gmsh
from pymor.discretizers.builtin.grids.oned import OnedGrid
from pymor.discretizers.builtin.grids.rect import RectGrid
from pymor.discretizers.builtin.grids.tria import TriaGrid



# This file is part of the pyMOR project (https://www.pymor.org).
# Copyright 2013-2021 pyMOR developers and contributors. All rights reserved.
# License: BSD 2-Clause License (https://opensource.org/licenses/BSD-2-Clause)

import numpy as np
import pytest
from hypothesis import assume, settings
from hypothesis.strategies import sampled_from

from pymor.algorithms.basic import almost_equal
from pymor.algorithms.svd_va import method_of_snapshots, qr_svd
from pymor.algorithms.basic import contains_zero_vector
from pymor.core.logger import log_levels
from pymor.vectorarrays.numpy import NumpyVectorSpace
from pymortests.base import runmodule
from pymortests.strategies import given_vector_arrays

methods = [method_of_snapshots, qr_svd]


@given_vector_arrays(method=sampled_from(methods))
@settings(deadline=None)
def test_method_of_snapshots(vector_array, method):
    A = vector_array

    # TODO assumption here masks a potential issue with the algorithm
    #      where it fails in internal lapack instead of a proper error
    assume(len(A) > 1 or A.dim > 1)
    assume(not contains_zero_vector(A, rtol=1e-13, atol=1e-13))

    B = A.copy()
    with log_levels({"pymor.algorithms": "ERROR"}):
        U, s, Vh = method(A, rtol=4e-8)  # default tolerance
    assert np.all(almost_equal(A, B))
    assert len(U) == len(s) == Vh.shape[0]
    assert Vh.shape[1] == len(A)
    assert np.allclose(Vh @ Vh.T.conj(), np.eye(len(s)))
    if len(s) > 0:
        U.scal(s)
        UsVh = U.lincomb(Vh.T)
        assert np.all(almost_equal(A, UsVh, atol=s[0]*4e-8*2))


@pytest.mark.parametrize('method', methods)
def test_method_of_snapshots_with_product(operator_with_arrays_and_products, method):
    _, _, A, _, p, _ = operator_with_arrays_and_products

    B = A.copy()
    with log_levels({"pymor.algorithms": "ERROR"}):
        U, s, Vh = method(A, product=p)
    assert np.all(almost_equal(A, B))
    assert len(U) == len(s) == Vh.shape[0]
    assert Vh.shape[1] == len(A)
    assert np.allclose(Vh @ Vh.T.conj(), np.eye(len(s)))
    U.scal(s)
    UsVh = U.lincomb(Vh.T)
    assert np.all(almost_equal(A, UsVh, rtol=4e-8))


@pytest.mark.parametrize('method', methods)
def test_not_too_many_modes(method):
    vec_array = NumpyVectorSpace.from_numpy(np.logspace(-5, 0, 10).reshape((-1, 1)))
    U, s, V = method(vec_array, atol=0, rtol=0)
    assert len(U) == len(s) == len(V) == 1


if __name__ == "__main__":
    runmodule(filename=__file__)
